(ns puppetlabs.classifier.schema
  (:require [clojure.set :refer [rename-keys]]
            [clojure.walk :as walk]
            [clj-time.format :as fmt-time]
            [fast-zip.core :as z]
            [fast-zip.visit :as zv]
            [schema.core :as sc]
            [slingshot.slingshot :refer [throw+]]
            [puppetlabs.kitchensink.core :refer [deep-merge mapvals parse-number]]
            [puppetlabs.classifier.storage :refer [root-group? root-group-uuid]]
            [puppetlabs.classifier.util :refer [map-delta uuid?]])
  (:import java.util.UUID))

;; Schemas
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(def ISO8601EncodedDateTime
  (sc/pred (fn [x]
             (let [iso-formatter (fmt-time/formatters :date-time-no-ms)]
               (and (string? x)
                    (not= :errored (try (fmt-time/unparse iso-formatter x)
                                     (catch Throwable _ ::errored))))))))

(def UUIDRepresentation
  (sc/pred #(and (or (string? %) (keyword? %))
                 (uuid? %))))

(def NodeField
  (sc/if #'sequential?
    [String]
    String))

(def NodeValue
  {:path NodeField
   :value sc/Any})

(def Environment
  {:name String})

(def PuppetClass ; so named to avoid clashing with java.lang.Class
  {:name String
   :parameters {sc/Keyword sc/Any}
   :environment String})

(def NoOpSigil
  {:environment String
   :no-op? boolean})

(defn no-op-sigil?
  [x]
  (and (map? x) (contains? x :no-op?)))

(def SynchronizationValue
  (sc/if no-op-sigil?
    NoOpSigil
    PuppetClass))

(def PuppetVariable #"^[a-zA-Z0-9_]+$")

(def PuppetVariableKeyword
  (sc/both sc/Keyword
           (sc/pred #(re-find PuppetVariable (name %)) "valid Puppet variable name")))

(def ENCValue
  (sc/conditional true? boolean
                  false? boolean
                  string? String
                  number? Number
                  map? {sc/Keyword (sc/recursive #'ENCValue)}
                  sequential? [(sc/recursive #'ENCValue)]))

(def ENCValueWithNilMapParamValues
  (sc/conditional true? boolean
                  false? boolean
                  string? String
                  number? Number
                  map? {sc/Keyword (sc/maybe (sc/recursive #'ENCValueWithNilMapParamValues))}
                  sequential? [(sc/recursive #'ENCValue)]))

(def RuleCondition
  (sc/conditional
    #(contains? #{"=" "~" "<" "<=" ">" ">="} (first %))
    [(sc/one (sc/enum "=" "~" "<" "<=" ">" ">=") "operator")
     (sc/one NodeField "field")
     (sc/one String "target-value")]

    #(= (first %) "and")
    [(sc/one (sc/eq "and") "conjunction")
     (sc/one (sc/recursive #'RuleCondition) "first-term") (sc/recursive #'RuleCondition)]

    #(= (first %) "or")
    [(sc/one (sc/eq "or") "disjunction")
     (sc/one (sc/recursive #'RuleCondition) "first-term") (sc/recursive #'RuleCondition)]

    #(= (first %) "not")
    [(sc/one (sc/eq "not") "negation") (sc/one (sc/recursive #'RuleCondition) "negated-expression")]))

(def ExplainedCondition
  {:value boolean
   :form (walk/prewalk
           (fn [form]
             (cond
               (and (= (type form) schema.core.Recursive)
                    (= (:schema-var form) #'RuleCondition))
               (sc/recursive #'ExplainedCondition)

               (= form (sc/one NodeField "field"))
               (sc/one NodeValue "field-with-value")

               :else form))
           RuleCondition)})

(def MeaningfulRuleCondition
  "This differs from RuleCondition only in that numeric comparison clauses must
  have a target value that can be parsed as a number."
  (let [numeric-comparison? #(contains? #{"<" ">" "<=" ">="} (first %))
        meaningful-comparison [(sc/one (sc/enum "<" ">" "<=" ">=") "comparison")
                               (sc/one NodeField "field")
                               (sc/one (sc/pred (comp number? parse-number)) "target-value")]
        eq-or-regex? #(contains? #{"=" "~"} (first %))
        eq-or-regex [(sc/one (sc/enum "=" "~") "operator")
                     (sc/one NodeField "field")
                     (sc/one String "target-value")]
        non-leaf-rule-clauses (drop 1 (:preds-and-schemas RuleCondition))]
    (-> non-leaf-rule-clauses
      (conj [numeric-comparison? meaningful-comparison])
      (conj [eq-or-regex? eq-or-regex])
      (->>
        (walk/prewalk (fn [form]
                        (if (and (= (type form) schema.core.Recursive)
                                 (= (:schema-var form) #'RuleCondition))
                          (sc/recursive #'MeaningfulRuleCondition)
                          form)))
        (hash-map :preds-and-schemas)
        sc/map->ConditionalSchema))))

(defn meaningful-rule?
  "Validates that a well-formed rule is also meaningful. Currently, that just
  means that numeric comparison operations have numeric target values."
  [rule]
  (nil? (sc/check MeaningfulRuleCondition rule)))

(defn meaningless-clauses
  "Finds the meaningless clauses in a RuleCondition. Returns a seq of said
  clauses, or nil if no meaningless clauses were found."
  [rule]
  (let [!clauses (atom [])]
    (walk/prewalk (fn [form]
                    (when (and (vector? form)
                               (contains? #{"<" ">" "<=" ">="} (nth form 0))
                               (nil? (parse-number (nth form 2))))
                      (swap! !clauses conj form))
                    form)
                  rule)
    (seq @!clauses)))


(def Rule
  {:when RuleCondition
   :group-id UUID
   (sc/optional-key :id) Number})

(def Classification
  {:environment String
   :environment-trumps boolean
   :classes {sc/Keyword (sc/maybe {sc/Keyword ENCValue})}
   :variables {PuppetVariableKeyword ENCValue}})

(def ClassificationOutput
  (dissoc Classification :environment-trumps))

(def ClassificationConflict
  {(sc/optional-key :environment) #{String}
   (sc/optional-key :classes) {sc/Keyword {sc/Keyword #{ENCValue}}}
   (sc/optional-key :variables) {PuppetVariableKeyword #{ENCValue}}})

(def CheckIn
  {:node String
   :time org.joda.time.DateTime
   :explanation {UUID ExplainedCondition}
   (sc/optional-key :classification) ClassificationOutput
   (sc/optional-key :transaction-uuid) UUID})

(def SerializableCheckIn
  (rename-keys CheckIn {:transaction-uuid :transaction_uuid}))

(def ClientCheckIn
  (assoc SerializableCheckIn
         :time ISO8601EncodedDateTime
         :explanation {UUIDRepresentation ExplainedCondition}
         (sc/optional-key :transaction_uuid) UUIDRepresentation))

(def Node
  {:name String
   :check-ins [(dissoc CheckIn :node)]})

(def SerializableNode
  {:name String
   :check_ins [(dissoc SerializableCheckIn :node)]})

(def ClientNode
  {:name String
   :check_ins [(dissoc ClientCheckIn :node)]})

(def SubmittedNode
  {:name String
   :fact {sc/Keyword sc/Any}
   :trusted {sc/Keyword sc/Any}})

(def Group
  (assoc Classification
         :name String
         :id UUID
         :parent UUID
         (sc/optional-key :rule) RuleCondition
         (sc/optional-key :description) String))

(def AnnotatedGroup
  (assoc Group
         (sc/optional-key :deleted) {sc/Keyword {:puppetlabs.classifier/deleted boolean
                                                 sc/Keyword {:puppetlabs.classifier/deleted boolean
                                                             :value ENCValue}}}))

(def SerializableGroup
  (rename-keys AnnotatedGroup {:environment-trumps :environment_trumps}))

(def ClientGroup
  (-> AnnotatedGroup
    (rename-keys {:environment-trumps :environment_trumps})
    (assoc :id UUIDRepresentation
           :parent UUIDRepresentation)))

(def GroupDelta
  {:id UUID
   (sc/optional-key :name) String
   (sc/optional-key :environment) String
   (sc/optional-key :environment-trumps) boolean
   (sc/optional-key :description) String
   (sc/optional-key :parent) UUID
   (sc/optional-key :rule) (sc/maybe RuleCondition)
   (sc/optional-key :classes) {sc/Keyword
                               (sc/maybe {sc/Keyword (sc/maybe ENCValueWithNilMapParamValues)})}
   (sc/optional-key :variables) {PuppetVariableKeyword (sc/maybe ENCValueWithNilMapParamValues)}})

(def ConflictDetails
  {:value ENCValue
   :from Group
   :defined-by Group})

(def ExplainedConflict
  {(sc/optional-key :environment) #{ConflictDetails}
   (sc/optional-key :classes) {sc/Keyword {sc/Keyword #{ConflictDetails}}}
   (sc/optional-key :variables) {PuppetVariableKeyword #{ConflictDetails}}})

(def SerializableExplainedConflict
  (-> ExplainedConflict
    (update-in [(sc/optional-key :environment)] vec)
    (update-in [(sc/optional-key :classes) sc/Keyword sc/Keyword] vec)
    (update-in [(sc/optional-key :variables) PuppetVariableKeyword] vec)))

(def ClientExplainedConflict
  (walk/prewalk (fn [form]
                  (if (= form UUID)
                    UUIDRepresentation
                    form))
                ExplainedConflict))

(def HierarchyNode
  {:group Group
   :children #{(sc/recursive #'HierarchyNode)}})

(def ValidationNode
  (assoc HierarchyNode
         :missing-referents (sc/maybe {sc/Keyword sc/Any})
         :unspecified-parameters (sc/maybe {sc/Keyword #{sc/Keyword}})
         :children (sc/maybe #{(sc/recursive #'ValidationNode)})))

;; Utilities for creating & converting maps conforming to the Schemas
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(sc/defn group->classification :- Classification
  [group :- Group]
  (select-keys group (keys Classification)))

(defn group->rule
  [group]
  {:when (:rule group) :group-id (:id group)})

(sc/defn group-delta :- GroupDelta
  "Returns a delta that, when applied, turns group `g` into group `h`"
  [g :- Group, h :- Group]
  {:pre [(= (:id g) (:id h))]}
  (assoc (map-delta g h)
         :id (:id g)))

(sc/defn unannotated :- Group
  [group :- AnnotatedGroup]
  (dissoc group :deleted))

(sc/defn group->serializable-group :- SerializableGroup
  [group :- (sc/if #(contains? % :deleted) AnnotatedGroup Group)]
  (rename-keys group {:environment-trumps :environment_trumps}))

(sc/defn node->serializable-node :- SerializableNode
  [node :- Node]
  (-> node
    (update-in [:check-ins] (partial map #(rename-keys % {:transaction-uuid :transaction_uuid})))
    (rename-keys {:check-ins :check_ins})))

(sc/defn explained-conflict->serializable-explained-conflict :- SerializableExplainedConflict
  "In order to produce JSON output that has a consistent ordering in the arrays
  that correspond to sets in Classification explanations, we need to just
  convert the sets to sorted arrays."
  [conflict :- ExplainedConflict]
  (let [sort-deets (partial sort-by (comp str :id :from))
        ->serializable-deets (comp vec sort-deets)]
    (cond-> conflict
      (contains? conflict :environment)
      (update-in [:environment] ->serializable-deets)

      (contains? conflict :classes)
      (update-in [:classes]
                 #(into {} (for [[class params] %]
                             [class (into {} (for [[param deets] params]
                                               [param (->serializable-deets deets)]))])))

      (contains? conflict :variables)
      (update-in [:variables] #(into {} (for [[variable deets] %]
                                          [variable (->serializable-deets deets)]))))))

(defn ->serializable
  [x]
  (if x
    (condp (comp nil? sc/check) x
      ;; This is a performance optimization -- currently (10/23/14) something
      ;; that matches Group will always match the AnnotatedGroup schema (since
      ;; the annotations are purely supplementary), so we just check for matches
      ;; on AnnotatedGroup in order to cut down on the total number of schema
      ;; checks. If annotations ever cease to be supplementary, such that
      ;; AnnotatedGroup is not a superset of Group, then this will need to
      ;; change.
      AnnotatedGroup (group->serializable-group x)
      [AnnotatedGroup] (map group->serializable-group x)
      Node (node->serializable-node x)
      [Node] (map node->serializable-node x)
      ExplainedConflict (explained-conflict->serializable-explained-conflict x)
      x)))

(defn- extract-cycle
  ([id id->parent-id id->group] (extract-cycle id id->parent-id id->group [id]))
  ([id id->parent-id id->group cycle]
   (let [last-id (or (last cycle) id)
         next-parent-id (get id->parent-id last-id)]
     (if (= next-parent-id id)
       (map id->group cycle)
       (recur id id->parent-id id->group (conj cycle next-parent-id))))))

(defn- children-by-id->parent-by-id
  [id->children]
  (reduce (fn [id->parent [id children]]
            (apply assoc id->parent (interleave (map :id children) (repeat id))))
          {}
          id->children))

(defn- groups->tree*
  [{id :id :as group} id->group id->children !marked]
  (when (get @!marked id)
    (throw+ {:kind :puppetlabs.classifier/inheritance-cycle
             :cycle (extract-cycle id (children-by-id->parent-by-id id->children) id->group)}))
  (let [children (if (root-group? group)
                   (remove root-group? (id->children id))
                   (id->children id))]
    (swap! !marked assoc id true)
    {:group group
     :children (set (map #(groups->tree* % id->group id->children !marked) children))}))

(sc/defn groups->tree :- HierarchyNode
  "Converts a flat collection of groups into a group hierarchy tree. The groups
  must form a complete, valid hierarchy wherein the root group has the expected
  id, all groups have the root as an ancestor, and there are no cycles. If any
  of these conditions are violated then an exception will be thrown."
  [groups :- [Group]]
  (when-not (apply distinct? (map :id groups))
    (let [duplicated-ids (for [[id groups] (group-by :id groups)
                               :when (> (count groups) 1)]
                           id)
          dupe-count (count duplicated-ids)]
      (throw+ {:kind :puppetlabs.classifier.storage/uniqueness-violation
               :entity-kind "group"
               :constraint "groups_pkey"
               :fields (repeat dupe-count "id")
               :values duplicated-ids})))
  (let [group-maps (loop [gs groups, id->g {}, id->c {}]
                     (if-not (seq gs)
                       [id->g id->c]
                       (let [g (first gs)]
                         (recur (next gs)
                                (assoc id->g (:id g) g)
                                (update-in id->c [(:parent g)] (fnil conj []) g)))))
        [id->group id->children] group-maps
        !marked (atom {})
        tree (groups->tree* (get id->group root-group-uuid) id->group id->children !marked)
        unreachable-groups (remove #(get @!marked (:id %)) groups)]
    (when (seq unreachable-groups)
      (throw+ {:kind :puppetlabs.classifier/unreachable-groups
               :groups unreachable-groups}))
    tree))

(sc/defn extract-classes :- [PuppetClass]
  [classifications :- [Classification]]
  (let [by-env (group-by :environment classifications)
        w-classes (mapvals (fn [clss8ns] (apply deep-merge (map :classes clss8ns))) by-env)]
    (for [[env classes] w-classes
          [class-kw params] classes]
      {:name (name class-kw)
       :environment (name env)
       :parameters (mapvals (constantly nil) params)})))

(defn- hierarchy-node?
  "Predicate that returns true if `x` conforms to either the HierarchyNode or
  ValidationNode schema."
  [x]
  (or (nil? (sc/check HierarchyNode x))
      (nil? (sc/check ValidationNode x))))

(def hierarchy-zipper
  (partial z/zipper hierarchy-node? (comp seq :children) (fn [n cs] (assoc n :children (set cs)))))

(defn tree->groups
  [root-node]
  (let [zip-root (hierarchy-zipper root-node)
        group-accum (zv/visitor
                      :pre [node groups]
                      {:state (conj groups (:group node))})]
    (:state (zv/visit zip-root [] [group-accum]))))
